#! /usr/bin/env python

#
# Copyright(C) 2005 INL
# Written by Jean Gillaux <jean@inl.fr>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#


__doc__ = """

This module is ressponsible for parsing xml nufw firewall desccription, and mapping it into objects.

""" 
import sys
import xml.dom.minidom
import IPy
import copy
from os import linesep
from net.defs import INTERNET

ZONE_EXT="external"
ZONE_INT="internal"
ZONE_VPN="vpn"

DFT_GATEWAY = "dftgateway"

class UnknownVpn(Exception):
    def __init__(self,mark):
	self.mark=mark
	
    def __str__(self):
	return 'unknown vpn with mark %s'%(self.mark)
    
class Net:
    """Maps <net> tag from a network description xml file (nuface interface)
    """
    def __init__(self,name,addr,zone='',type='',mark='',remote='', id=None):
	if(zone==''):
	    self.__zone=ZONE_INT
	else:
	    self.__zone=zone
	self.__conns=[]
	if addr == '': addr = '0.0.0.0/0'
	self.__addr=IPy.IP(addr)
	self.__name=name
	self.__type=type
	self.__mark=mark
	self.__remote = remote
	self.__id = 0
	if id is not None: self.__id = id
	
    def add_conn(self,dir,fwid,iface,gw=None, dftgw=None):
	h={'direct':dir, 'fwid':fwid, 'iface':iface}
	#if(len(args)==1):
	#    h['gateway']=args[0]
        if gw:
            h['gateway']=gw
        if dftgw:
            h[DFT_GATEWAY]=dftgw
	#print h
	self.__conns.append(h)
	
    def get_conn(self,fwid):
	for conn in self.__conns:
	    if conn['fwid']==fwid:
		return conn
	
    def iter_conns(self):
	for index in range(0,len(self.__conns)):
	    yield self.__conns[index]
    
    #getters
    def netmask(self):
	"""get the netmask of the network
	"""
	return self.__addr.netmask()

    def ip(self):
	"""return the string representing the addr of the Net object"""
	return str(self.__addr)

    def addr(self):
	return self.__addr
    
    def name(self):
	return self.__name
   
    def zone(self):
	return self.__zone
	
    def type(self):
	return self.__type
    
    def mark(self):
	return self.__mark
   
    def remote(self):
	return self.__remote
   
   
    def is_vpn(self):
	"""Checks if the network is a vpn """
	return self.__mark != ''
   
    def id(self):
	"""Get network id"""
	return self.__id
   
    def __str__(self):
	s='name: %s, addr:%s, id:%s, zone:%s' %(self.__name,str(self.__addr), str(self.__id), self.__zone)
	return s
    def __repr__(self):
	return str(self)

    def __eq__(self, other):
        #print "in Net.__cmp__ 1", self, other
        if self.__id == other.__id and self.__addr == other.__addr:
            #print "in Net.__cmp__ ", self, other
            #print "in Net.__cmp__  return True"
            return True
        #print "in Net.__cmp__ 1 return False"
        return False
    def __ne_(self_other):
        return self.__eq__(other)
#End class Net

class Interface:
    """An interface has an ip addresse and a list of networks attached to it.
    """
    def __init__(self,id,name,addr):
	self.__id=id
	self.__net=None
	self.__addr=IPy.IP(addr)
	self.__name=name
	self.__distnets=[]
	self.__nets=[]
	self.__gateways = {}
	
    def name(self) :
	return self.__name

    def set_net(self,net):
	self.__net=net
	
    def add_distnet(self,net):
	self.__distnets.append(net)
    	
    def add_net(self,net):
	self.__nets.append(net)
    	#if 
    
    def distnets(self):
	for n in self.__distnets:
	    yield n
    
    
    
    def allnets(self):
	"""return all the network of the interface
	"""
	for n in self.__nets:
	    yield n
    
    def netmask(self):
	if(type(self.__net)==type(None)):
	    return ''
	return self.__net.netmask()

    def addr(self):
	return str(self.__addr)
    def addr2(self):
        return self.__addr
   
    def id(self):
	return self.__id
    
    def net(self):
	return self.__net
    def self_net(self):
	"""return a Net object that represents the ip addr of the interface"""
	return Net(self.__name,str(self.__addr))
	pass
    def deb_conf(self):
	s="""
auto %s
iface %s inet static
    address %s
    netmask %s""" %(self.__name, self.__name, self.__addr,self.netmask())
	return s

    
#End class Interface
class RuleList(object):
    """List of rules
    """
    def __init__(self):
	self.rules = []
	self.bad = 0

    def is_bad(self):
	return self.bad == 1
    
    def add_rule(self,rule):
	self.rules.append(rule)
	
class FWNetConf(object):
    """Dns configuration for firewall
    """
    def __init__(self):
	self.__dns_search = []
	self.__dns_nameservers = []
	pass

    def add_dns_search(self,resolv):
	"""add a search item for dns configuration.
	parameters: resolv = str
	"""
	self.__dns_search.append(resolv)
    
    def add_dns_nameserver(self,server):
	"""Add a dns namserver.
	parameters: server = str. the ip addr of the namserver
	"""
	self.__dns_nameservers.append(server)

    def get_dns_search(self):
	"""Return list of dns resolv domains
	"""
	return self.__dns_search
    def get_dns_nameservers(self):
	"""Return the list of dns nameservers
	"""
	return self.__dns_nameservers
    
    def get_dns_conf(self):
	"""Get dns configuration. Returns (resolv list, nameservers list)
	"""
	return (self.__dns_search, self.__dns_nameservers)
#End class FWNetConf
    
class FW:
    """FW object is a firewall with:
- Interfaces
- Nets
- Rules
Rules are stored in chains corresponding to their source/destination ip addresses.
    """
    def __init__(self,id):
	self.__ifaces=[]
	self.__nets=[]
	self.__id=id
	self.__hrules={}
	self.__allrules=[]
        self.__nat_rules = {'snat': [], 'dnat': [], 'pnat': []}
	self.__chains=[]
	self.__ext_net=None
	self.__vpns={}
	self.manage_input = True
	self.manage_output = True
	self.manage_forward = True
	self.__netconf = None
        self.queue = None
        self.auth_ext = False #authenticate packets from internet?

    def has_dft_net(self):
        return self.__ext_net is not None

    def add_iface(self,iface):
	self.__ifaces.append(iface)
	
    def ifaces(self):
	for i in self.__ifaces:
	    yield i

    def add_net(self,net, conn=None):
        #print "in fw.add_net: ",conn
	self.__nets.append(net)
	#self.__chains=self.chains(all=1)
	#if net.zone()==ZONE_EXT: #TODO: test & remove
	#    self.__ext_net=net
	#More logical than net.zone():
	if conn is not None and conn.has_key(DFT_GATEWAY):
            #print 'setting internet'
	    self.__ext_net = net
	if net.type()=='mark':
	    self.__vpns[net.mark()]=net
	
	  
    #  Checks if a network is the default route network for the firewall
    def is_default_net(self, net):
	"""Checks if a network is the default route network for the firewall
	   net: Net object, the network to be checked
	"""
        if not self.__ext_net:
            return False
	return net.id() == self.__ext_net.id()
    
    def is_normal_net(self,net):
	"""Check if a network is not default route nor vpn"""
	return (net.is_vpn() is not True and self.is_default_net(net) is not True)
    
    def deb_conf(self):
	"""Returns a debian /etc/interfaces network configuration
	"""
	s = ''
	for myiface in self.__ifaces:
	    s+="""%s
    %s""" %(myiface.deb_conf(),self.debconf_gateways(myiface))
	return s
	
    def debconf_gateways(self,iface):
	"""get 'route add' commands for network behind interface, for inclusion in /etc/network/interface post-up sections
	"""
	s = ''
	n = iface.net()
	if n:
	    h = n.get_conn(self.__id)
	    if h.has_key(DFT_GATEWAY):
	        s += 'gateway %s%s'%(h[DFT_GATEWAY],linesep)
	for n in iface.distnets():
	    h = n.get_conn(self.__id)
	    #if h.has_key('gateway') and n.zone() not in [ZONE_VPN]: #do not use zones anymore
	    print h
	    if h.has_key(DFT_GATEWAY):
	    	s += 'gateway %s%s'%(h[DFT_GATEWAY],linesep)
	    if h.has_key('gateway') and not n.is_vpn():
		s += 'post-up route add -net %s gw %s%s'%(n.ip(),h['gateway'],linesep)
	return s
    
    def routes(self):
	s=''
	for iface in self.__ifaces:
	    for n in iface.distnets():
		h=n.get_conn(self.__id)
		#print 'route ',h
		s+="%s\t%s\n" %(str(n.addr().net()),h['router'])
	return s

    def __add_chain_for_rules_if_not_exist(self,chain):
	if not self.__hrules.has_key(chain):
	    self.__hrules[chain] = RuleList()
    
    def find_zone(self,ip,mark):
	"""Find the zone of the firewall which contains the string ip address "ip" and/or 
	the string mark
	"""
	if mark != '' and self.__vpns.has_key(mark):
	    #print 'zone vpn'
	    return ZONE_VPN
	for net in self.__nets:
	    if net.zone != ZONE_EXT and ip in net.addr():
		#print 'net zone ',net.zone(),ip
		return net.zone()
	#print 'zone external',ip
	return ZONE_EXT
	
    def is_correct_network(self, network):
        """Check if network is correct
        network should be included in one enterprise network
        if not: network must not overlap with enterprise network
        """
        net = IPy.IP(network)
        l = [IPy.IP(n) for n in self.entreprise_nets()]
        included = 0
        overlap = 0
        for net2 in l:
            if net in net2:
                included = 1
        if included: 
            return True
        else:
            for nete in l:
                if nete.overlaps(net):
                    return False
            return True
            
            
        
    def chain(self,ip_src,ip_dst,zone_src='',zone_dst='',mark_src='',mark_dst=''):
	"""Returns the fw chain that fits ip_src/ip_dst/zone_src/zone_dst/mark
	"""
	# FIXME / TODO
	# do not use anymore zone attribute for elements. (OK for networks)
	chp1 = None
	chp2 = None
        if1 = None
        if2 = None
	find1=0
	find2=0
	#print 'in chain |%s| |%s| |%s| |%s|' %(ip_src,zone_src,zone_dst,ip_dst)
	if mark_src!='': #Find VPN
	    #print 'chain mar_src', mark_src
	    if not self.__vpns.has_key(mark_src):
		raise UnknownVpn(mark_src)
	    chp1=self.__vpns[mark_src]
	    find1=1
	if mark_dst!='':
	    #print 'chain mar_src', mark_dst
	    if not self.__vpns.has_key(mark_dst):
		raise UnknownVpn(mark_dst)
	    chp2=self.__vpns[mark_dst]
	    find2=1

	#Let's find the right network. no need of rules zones.
	for iface in self.__ifaces:
	    for net in iface.allnets():
		if not self.is_default_net(net): #we don't want to match default net for now: it comes at last
		    if ip_src in net.addr():
			chp1 = net
			find1 = 1
		    if ip_dst in net.addr():
			chp2 = net
			find2 = 1
	    #INPUT or OUTPUT: all interfaces, default included
	    if ip_src == iface.addr():
		chp1 = Net('IF',iface.addr(),type='ipv4',zone=ZONE_INT)
                if1 = iface
		find1 = 1
	    if ip_dst == iface.addr():
		chp2 = Net('IF',iface.addr(),type='ipv4',zone=ZONE_INT)
                if2 = iface
		find2 = 1
	#have we find what we want? if not, got to the evil default route.
        #FIXME NOINT what todo if no ext_net?
	if not chp1:
	    chp1 = self.__ext_net
	    find1 = 1
	if not chp2:
	    chp2 = self.__ext_net
	    find2 = 1
	if not if1:
            if1 = self.get_iface_by_net(chp1)
        if not if2:
            if2 = self.get_iface_by_net(chp2)
	#print chp1.name(), chp2.name()
	return {'chain':chp1.name()+'-'+chp2.name(), 'netsrc':chp1, 'netdst':chp2, 'ifsrc': if1, 'ifdst': if2}
	
    def chains(self,all=0,inout=0):#TODO: get INPUT:OUTPUT chains with inout=1
	"""returns a list of firewall's chains.
all=1 : all networks are considered , including distant ones
inout = 1: chains for input/output are also generated
	"""
	l = []
	l2 = []
	for iface in self.__ifaces:
	    #print (all==0 and iface.net().zone()!=ZONE_EXT) or all==1
	    #if((all==0 and iface.net().zone()==ZONE_INT) or all==1):
	    if( (all==0 and self.is_normal_net(iface.net())) or all==1):
		l.append({'name':iface.net().name(), 'net':iface.net(),'zone':iface.net().zone(),'iface':iface})
	    for n in iface.distnets():
		if((all==0 and self.is_normal_net(iface.net())) or all==1):
		    l.append({'name':n.name(), 'net':n,'zone':iface.net().zone(),'iface':iface})
	    if inout == 1:
		choutput = {'name':'IF-'+iface.net().name(),'net_src':iface.self_net(),'net_dst':iface.net(),'if_src':iface,'if_dst':iface}
		self.__add_chain_for_rules_if_not_exist('IF-'+iface.net().name())
		chinput = {'name':iface.net().name()+'-IF','net_src':iface.net(),'net_dst':iface.self_net(),'if_src':iface,'if_dst':iface}
		self.__add_chain_for_rules_if_not_exist('%s-IF'%(iface.net().name()))
		l2.extend([choutput,chinput])
		for n in iface.distnets():
		    choutput = {'name':'IF-'+n.name(),'net_src':iface.self_net(),'net_dst':n,'if_src':iface,'if_dst':iface}
		    self.__add_chain_for_rules_if_not_exist('IF-'+n.name())
		    chinput = {'name':n.name()+'-IF','net_src':n,'net_dst':iface.self_net(),'if_src':iface,'if_dst':iface}
		    self.__add_chain_for_rules_if_not_exist('%s-IF'%(n.name()))
		    l2.extend([choutput,chinput])

		#l.append({'name':iface.name(), 'net':iface.self_net(),'zone':iface.net().zone(),'iface':iface})
	#print "in chains ",l
	def _fun(l1,l2):
	    res=[]
	    for u1 in l1:
		for u2 in l2:
		    if(u1['name']!=u2['name']):
			res.append({'name':u1['name']+'-'+u2['name'], 'net_src':u1['net'], 'net_dst':u2['net'],'if_src':u1['iface'],'if_dst':u2['iface']})
			if not self.__hrules.has_key(u1['name']+'-'+u2['name']):
			    self.__hrules[u1['name']+'-'+u2['name']] = RuleList()
	    return res
	tmpres = _fun(l,l)
	tmpres.extend(l2)
	return tmpres

    	
    def entreprise_nets(self):
	l=[]
	for iface in self.__ifaces:
	    n=iface.net()
	    #if(n.zone()==ZONE_INT): #no more zone
	    if self.is_normal_net(n):
		l.append(str(n.addr()))
	    for n in iface.distnets():
		#if(n.zone()==ZONE_INT): #no more zone
		if self.is_normal_net(n):
		    l.append(str(n.addr()))
	return l

    def add_rule(self, r):
	"""Add a rule to the firewall into the right chain
	"""
	#print 'in add_rule',r.aclid,r.elt_src(), r.elt_dest()
	#if r.is_auto == 1: print 'rule is auto',r
	elt_src = r.elt_src()
	elt_dst = r.elt_dest()#find and set the zones if they're not set
        netsrc = IPy.IP(elt_src.get('numnet'))
        netdst = IPy.IP(elt_dst.get('numnet'))
        # check inclusion and overlaping for dest/src networks with enterprise networks
        if netsrc != INTERNET and not self.is_correct_network(elt_src.get('numnet')):
            raise ValueError("network not correct: %s"%(elt_src.get('numnet')))
        if netdst != INTERNET and not self.is_correct_network(elt_dst.get('numnet')):
            raise ValueError("network not correct: %s"%(elt_dst.get('numnet')))
	ch=self.chain(r.elt_src().get('numnet'),r.elt_dest().get('numnet'),zone_src=r.elt_src().get('zone'),zone_dst=r.elt_dest().get('zone'),mark_src=r.elt_src().get('mark'),mark_dst=r.elt_dest().get('mark'))
	r.set_netsrc(ch['netsrc'])
	r.set_netdst(ch['netdst'])
	r.snetsrc=str(ch['netsrc'].addr())
	r.snetdst=str(ch['netdst'].addr())
	#print 'in add_rule ', ch, r.elt_src(), r.elt_dest()
	if(not self.__hrules.has_key(ch['chain'])):
	    self.__hrules[ch['chain']] = RuleList()
	    self.__hrules[ch['chain']].bad = 1
	    #print "%s is bad"%(ch['chain'])
	#print 'rule added', r, ch['chain']
	self.__hrules[ch['chain']].add_rule(r)
	self.__allrules.append(r)
  
    def set_nat_rules(self, snat,dnat, pnat):
        """Set NAT rules for FW object
        """
        # set rules interfaces
        if self.__ext_net:
            ext_iface = self.get_iface_by_net(self.__ext_net).name()
        for r in snat:
            d = self.chain(r._to.strNormal(),r._to.strNormal())
            r._iface = d['ifdst'].name()
        # nat/pnat: see if destination address is set
        for r in dnat:
            if r._src:
                d = self.chain(r._src.strNormal(),r._src.strNormal())
                r._iface = d['ifdst'].name()
            else:
                r._iface = ext_iface
        for r in pnat:
            if r._dst:
                d = self.chain(r._dst.strNormal(),r._dst.strNormal())
                r._iface = d['ifdst'].name()
            #print r._iface
        self.__nat_rules = {'snat': snat, 'dnat': dnat, 'pnat': pnat}
        
    def find_network(self, addr):
	"""Given a ip address, find the network that this ip belongs to"""
	for iface in self.__ifaces:
	    for net in iface.allnets():
		if not self.is_default_net(net): 
		    if addr in net.addr():
			return net
	return self.__ext_net#FIXME NOINT
    
    def rule_dst_is_src(self,r):
	"""Test if a rule comes an goes to the same network """
	src = r.elt_src(); net_src = None
	dst = r.elt_dest(); net_dst = None
	if src.get('mark') != '': net_src = self.__vpns[src.get('mark')]
	if dst.get('mark') != '': net_dst = self.__vpns[src.get('mark')]
	if net_src is None:
	    net_src = self.find_network(src.get('numnet'))
	if net_dst is None:
	    net_dst = self.find_network(dst.get('numnet'))
	if self.is_default_net(net_src) and self.is_default_net(net_dst): return True
	if net_dst.addr() in net_src.addr() and net_src.addr() in net_dst.addr(): return True
	return False
	
   
    def rule_inout(self,rule):
	"""Test if the rule has its src/dest = one of the fw ifaces"""
	return self.rule_input(rule) or self.rule_output(rule)
    
    def rule_input(self,rule):
	"""Test if the rule has its dest = one of the fw ifaces"""
	iptest = rule.elt_dest().get('numnet')
	for iface in self.__ifaces:
	    if iface.addr() == iptest: return True
	return False

    def rule_output(self,rule):
	"""Test if the rule has its source = one of the fw ifaces"""
	iptest = rule.elt_src().get('numnet')
	for iface in self.__ifaces:
	    if iface.addr() == iptest: return True
	return False
    def chain_input(self, chain):
    	"""Test if the chain has its dest = one of the fw iface address"""
	return self.is_iface_address(chain['net_dst'].ip())
	
    def chain_output(self,chain):
	"""test if the chain has its source = one of the fw iface address"""
	#return self.is_iface_address(chain['net_src'].ip())
	if self.is_iface_address(chain['net_src'].ip()):
	    #print 'chain %s is output %s' %(chain,chain['net_src'].ip())
	    return True
	return False
	
    def chain_inout(self,chain):
	"""test if the chain has its src or dst = one of the fw ifaces address"""
	return self.chain_input(chain) or self.chain_output(chain)

    def is_iface_address(self, addr):
	"""Test if the string addr is one of the fw ifaces addresses"""
	for iface in self.__ifaces:
	    #print "is_iface_address: %s <--> %s"%(addr,iface.addr())
	    if iface.addr() == addr: return True
	return False

    def hrules(self):
	return self.__hrules

    def allrules(self):
	return self.__allrules

    def get_nat_rules(self):
        return [self.__nat_rules['snat'], self.__nat_rules['dnat'], self.__nat_rules['pnat']]

    def set_netconf(self,net_conf):
	"""set netconf member of FW object
	"""
	self.__netconf = net_conf

    def get_netconf(self):
	"""Get Net Conf of Firewall
	returns: FwNetConf object
	"""
	return self.__netconf
        
    def get_iface_by_net(self, net):
        """ Get iface corresponding to a specific network"""
        for tmpiface in self.__ifaces:
            if tmpiface.net() == net:
                return tmpiface
        return None
    
    def get_ext_net(self):
        return self.__ext_net
#End class FW


def _xml_parse_fwnetconf(xmlnode):
    """returns a FWNetConf object
    parameters: xmlnode = XmlNode (fw node)
    """
    ret = FWNetConf()
    dns = xmlnode.getElementsByTagName('dns')
    if not dns: return None
    for mynode in dns[0].childNodes:
	if mynode.nodeType == xml.dom.minidom.Node.ELEMENT_NODE and mynode.nodeName == 'search':
	    if mynode.hasAttribute('val'):
		ret.add_dns_search(mynode.getAttribute('val'))
	elif mynode.nodeType == xml.dom.minidom.Node.ELEMENT_NODE and mynode.nodeName == 'nameserver':
	    if mynode.hasAttribute('val'):
		ret.add_dns_nameserver(mynode.getAttribute('val'))
    return ret


def parse(file):
    """Parse a file that follows nuface firewall description xml format.
Returns a list of FW objects.
    """
    fwhash={}
    doc=xml.dom.minidom.parse(file)
    fwl=[]
    netl={}
    eltsfw=doc.getElementsByTagName("fw")
    eltnets=doc.getElementsByTagName('nets')
    eltsnet=eltnets[0].childNodes
    for eltnet in eltsnet:
	if(eltnet.nodeType==xml.dom.minidom.Node.ELEMENT_NODE and eltnet.nodeName=="net"):
	    cnet=Net(eltnet.getAttribute('name'),eltnet.getAttribute('addr'),zone=None,type=eltnet.getAttribute('type'),mark=eltnet.getAttribute('mark'),remote=eltnet.getAttribute('remote'), id=eltnet.getAttribute('id'))
	    if(eltnet.hasChildNodes()):
		#print "\thas childs"
		for conn in eltnet.getElementsByTagName('connection'):
		    fwid=conn.getAttribute('fwid')
		    direct=conn.getAttribute('direct')
		    iface=conn.getAttribute('iface')
		    #sd=''
		    if(direct=='1'): 
			#sd='directe'
			cnet.add_conn(direct,fwid,iface,dftgw=conn.getAttribute(DFT_GATEWAY))
		    else: 
			#sd='indirecte'
			cnet.add_conn(direct,fwid,iface,gw=conn.getAttribute("gateway"))
		#put new found net into global nets list
		    if(not netl.has_key(fwid)): netl[fwid]=[cnet]
		    else: netl[fwid].append(cnet)
    for elt in eltsfw:
	#print elt
	fwid=elt.getAttribute('id')
	eltsinterface=elt.getElementsByTagName('interface')
	myfw=FW(fwid)
        # set queue number (NFQUEUE)
        if elt.hasAttribute('queue'):
            queue = elt.getAttribute('queue')
            if queue.isdigit():
                myfw.queue = queue
	myfw.set_netconf(_xml_parse_fwnetconf(elt))#get net conf attributes
	for interface in eltsinterface:
		myiface=Interface(interface.getAttribute('id'),interface.getAttribute("name"),interface.getAttribute('addr'))
		myfw.add_iface(myiface)
		for net in netl[fwid]:
		    for conn in net.iter_conns():
		    	#print "\tconn ",conn
			if (conn['fwid']==fwid and conn['iface']==myiface.id()):
			    if conn['direct']=='1':
				myiface.set_net(net)
			    else: 
				#print "adding ",conn
				myiface.add_distnet(net)
			    myiface.add_net(net)
			    break # TODO: we have the connection
		    myfw.add_net(net, conn=conn)
		#print myiface.deb_conf()
	fwhash[fwid]=myfw
    return fwhash

    
