#! /usr/bin/env python

#
# Copyright(C) 2005 INL
# Written by Jean Gillaux <jean@inl.fr>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#


import copy
import socket
import IPy
from sets import Set
import string

#import ipt
class Error(ValueError):
    """Base class for exceptions
    """
    pass

class LoopError(Error):
    """Exception raised when a loop is detected in links
    """
    def __init__(self, value):
	self.value = value
	
    def __str__(self):
        return repr(self.value)

class LinkError(Error):
    """Exception raised when an link to a non-existent ressource or subject
    """
    def __init__(self, value):
	self.value = value
	
    def __str__(self):
        return repr(self.value)


def proto_number(proto):
    if proto == 'tcp':
	return 6
    if proto == 'udp':
	return 17
    if proto == 'icmp':
	return 1
class Elt:
    ID=''
    def __init__(self,*args):
	#self.ID=s
	if(len(args)==1):
	    self.__hash=args[0]
	else:
	    self.__hash={}
	self.__hash['id_son'] = '0'

    def get(self,key):
	"""Get value of element attibute designed by <key>
	   empty chain is returned if the attribute doesn't exist
	"""
	return self.__hash.get(key,'').encode("utf-8")
    
    def set(self,key,value):
	"""Set the key attribute element with value
	"""
    	self.__hash[key]=value
    
    def keys(self):
	return self.__hash.keys()
    
    def values(self):
	return self.__hash.values()

    def merge_elt(self,other):
	"""merge attributes of other element into self element
	"""
	for k in other.keys():
	    so=other.get(k)
	    sv=self.get(k)
	    if (sv==''):
		self.set(k,so)
	    #else:
		#sv=sv+so
		#self.set(k,sv)
    
    def __repr__(self):
	s="Elt: "
	for k in self.__hash.keys():
	    #print "__hash[k] ",self.__hash[k]
	    s+=k+"="+str(self.__hash[k])+" "
	return s
    def __cmp__(self,other):
	return cmp(self.__hash,other.__hash)
	

class EltSet:
    """A set of elements with <and> operator
    """
    def __init__(self,l):
	self.elts=l

    def add_elt(self,e):
	self.elts.append(e)
    
    def iter(self):
	for i in range(len(elts)-1,-1,-1):
	    yield(elts[i])
  
    def merge(self,other):
	"""Merge other's elements into self
	"""
	for el in other.elts:
	    try:
		ind = self.l.index(el)
		self.add_elt(el)
	    except ValueError:
		pass
    def __repr__(self):
	s="{"
	for elt in self.elts:
	    s+=elt.__repr__()+"| "
	s+="}"
	return s

    def merge_elts(self):
	if(self.elts!=[]):
	    e=self.elts[0]
	    for o in self.elts[1:len(self.elts)]:
		#print "in merge_elts", e,o
		e.merge_elt(o)
		#print "in merge_elts",e
	    self.elts=[e]
#end EltSet

class EltGrp:
    """Represents a Ressource or a subject
    """
    def __init__(self,ID,name,op):
	self.ID=ID
	self.op=op
	self.name=name
	self.link_t=[]
	self.eltset_t=[]
	self.eltgrp_t=[]
	self.elts=[]
	self.scpt=0

    #def __str__(self):
	#return str(self.eltset_t) + str(self.elts)
    
    def add_eltset(self,eltset):
	self.eltset_t.append(eltset)

    def remove_eltset(self,eltset):
	self.eltset_t.remove(eltset)
    
    def add_eltgrp(self,eltgrp):
	self.elt_grp.append(eltgrp)
    
    def add_elt(self,e):
	self.elts.append(e)

    def add_eltlink(self,lk):
	self.link_t.append(lk)
    
    def iter_eltsets(self):
	for ind in range(len(self.eltset_t)-1,-1,-1):
	    yield self.eltset_t[ind]
    
    def has_link(self):
	if(self.link_t==[]):
	    return False
	return True

    def is_ok(self):
	for es in (self.link_t):
	    for elt in es.elts:
		#print "top ",elt
		if(elt.get('type')=="link"):
		    return False
	return True
    
    def is_solved(self):
	"""Tells if the group of elements is solved
	"""
	if(self.eltgrp_t == []):
	    return True
	return False
	
    def resolv_or(self):
	"""Put elements that come from links with or operator in elements list
	"""
	#print 'resolv or eltgrp_t ',self.ID,self.eltgrp_t
	if(self.eltset_t==[]):
	    if(self.eltgrp_t!=[]):
		self.eltset_t=copy.deepcopy(self.eltgrp_t[0].eltset_t)
		#del(self.eltgrp_t[0])
		tmp=self.eltgrp_t[0]
		self.eltgrp_t.remove(tmp)
		#print 'resolv or eltgrp_t vide',self.ID,self.eltgrp_t

	#print "fffff ",self.eltset_t
	for grp in self.eltgrp_t:
	    if(len(self.eltset_t)==0):
		#print "resolv_or: eltset_t vide"
		pass
	    #print "len eltset_t: ",len(self.eltset_t), self.eltset_t
	    #if(self.eltset_t==[]):
		#print "pouetpou"
	    #print "\tin resolv_or",grp
	    if(grp.op=="or"):
		#print "\tin resolv_or",grp
	    #2 cases switch self.op
		if(self.op=="or"):
		    #print "resolv_or: self.op=or", self.ID
		    for eltset in grp.iter_eltsets():
			#print '\t\taddt eltset: ',self.ID,eltset.elts
			self.add_eltset(eltset)
		elif(self.op=="and"):
		    tmplist=[]
		    #print "resolv_or: self.op=and",self.ID
		    for myeltset in self.eltset_t:
			#print "\tmyeltset ",self.ID,myeltset.elts
			for eltset in grp.eltset_t:
			    if(eltset.elts!=[]):
				#for elt in eltset.elts:
				clone=EltSet(copy.copy(myeltset.elts))
				#print '\t elts to be added ',eltset.elts
				for elt in eltset.elts:
				    #print "eltset: ", eltset.elts
				    #print "puet", elt, eltset.elts
				    #print "clone ",clone.elts
				    clone.add_elt(elt)
				    #print "titi"
				    #print "myeltset ",myeltset.elts
				
				tmplist.append(clone)
				#print "\tClone added",self.ID,clone.elts
			#print "toto"
		    if(tmplist!=[]):
			self.eltset_t=tmplist
	self.__del_links("or")
	self.op="or"
    #End resolv_or
    
    def resolv_and(self):
	"""Put elements that come from links with or operator in elements list
	"""
	#print "in resolv_and"
	for grp in self.eltgrp_t:
	    if(grp.op=="and"):
		#print "resolv_and ", self.ID,grp
		if(self.op=="and"):
		    for myeltset in self.eltset_t:
			for eltset in grp.eltset_t:
			    #print 'in resolv and merge', self.ID,myeltset,eltset
			    myeltset.merge(eltset)
		elif(self.op=="or"):
		    for eltset in grp.eltset_t:
			#print 'in resolv and add',self.ID, eltset
			self.add_eltset(eltset)
	self.__del_links("and")
	#print "end resolv_end"
    #End __resolv_conf
    
    def resolv(self):
	"""Resolvs grpelts with <and> and <or> operators
	   Then clean links groups
	"""
	
	#if(self.eltgrp_t==[]):
	    #print 'begin resolv: eltgrp_t=[]',self.ID,self.scpt
	#else:
	    #print 'begin resolv: eltgrp_t!=[]',self.ID,self.scpt
	self.resolv_or()
	self.resolv_and()
	#print 'end resolv eltset_t ', self.ID,self.eltset_t
	self.eltgrp_t=[]
	self.scpt+=1
	#if(self.eltgrp_t==[]):
	    #print 'end resolv: eltgrp_t=[]',self.ID,self.scpt
	#else:
	    #print 'end resolv: eltgrp_t!=[]',self.ID,self.scpt
	

    def base_norm(self,grplist,linklist):
	"""Follow links elts with deep 1 and put the destination grpelts
	"""
	for es in (self.link_t):
	    for elt in es.elts:
		#print "base norm entering subloop"
		if(elt.get('type')=="link"):
		    dest=elt.get('linkid')
		    id=elt.get('ID')
		    grpdest=grplist.find(dest)
		    if grpdest is None: raise LinkError, "Link Error %s" %(dest)
		    #print "in base norm: ",grpdest
		    #print "dest : ",self.ID,self.name,grpdest.ID, grpdest.name
		    gid=grpdest.ID
		    if not linklist.has_key(self.ID):
			linklist[self.ID]=Set([gid])
		    else:
			#if gid not in linklist[self.ID]:
			linklist[self.ID].add(gid)
		    #print "in base_norm, linklist : " ,linklist,id,gid
		    if linklist.has_key(gid):
			if self.ID in linklist[gid]:
			    print linklist[gid]
			    raise LoopError('Link Loop Detected in grp ID %s and elt %s: dest %s already seen' %(self.ID,id,gid))
		    self.eltgrp_t.append(grpdest)
	if linklist.has_key(self.ID):
	    l=[]
	    for k in linklist.keys():
		lrem=linklist[self.ID]
		if self.ID in linklist[k]:
		    linklist[k].remove(self.ID)
		    linklist[k] |= lrem
		
	    
		    #linklist.append(gid)
	#self.link_t=[]

    def norm(self,grplist,linklist,listsolved=[]):
	"""The recursive thing
	"""
	#print 'Debut norm ',self.ID, listsolved
	#print "entering  norm()"
	if(self.has_link()):
	    #print "norm: base_norm",self
	    self.base_norm(grplist,linklist)
	    for sub in self.eltgrp_t:
		#print 'yrp', sub
		sub.norm(grplist,linklist,listsolved=listsolved)
	#print "norm part 2"	
	if(not self.is_solved()):
	    for foo in self.eltgrp_t:
		#print 'in norm not solved',self.ID,foo.ID
		#foo.resolv()
		pass
	    if listsolved.count(self.ID)==0:
		self.resolv()
	    listsolved.append(self.ID)
	#print 'End norm ',self.ID
	
    def __del_links(self,op):
	"""delete GrpEltSets in links list wich have <op> operator
	"""
	for grp in self.eltgrp_t:
	    if grp.op==op:
		self.eltgrp_t.remove(grp)
	    
	
    def __repr__(self):
	s=""
	s+="EltGrp: <"+self.ID + "> " + "<" + self.name + "> " + "<" + self.op + ">\n"
	for es in self.eltset_t:
	    s+="\tEltSet <"+es.__repr__()+">\n"
	return s

    def make_graph(self):
	#print 'digraph G {'
	for item in self.link_t:
	    for elt in item.elts:
		#print '%s -> %s' %(self.ID,elt.get('linkid'))
		print '%s -> "elt_%s"' %(self.ID,elt.get('ID'))
		print '"elt_%s" -> %s' %(elt.get('ID'),elt.get('linkid'))
	for set in self.eltset_t:
	    for elt in set.elts:
		print 'node [shape=box]; "elt_%s";'%(elt.get('ID'))
		print '%s -> "elt_%s"' %(self.ID,elt.get('ID'))
	#print '}'
#end class EltSet

class EltGrpList:
    def __init__(self):
	self.__list=[]

    def add(self,eltgrp):
	self.__list.append(eltgrp)

    def find(self,index):
	for a in self.__list:
	    if (a.ID==index):
		return a
	return None
    def normall(self,listsolved=[]):
	l=[]
	l2={}
	for grp in self.__list:
	    grp.norm(self,l2,listsolved=l)

    def make_graph(self):
	print 'digraph G {'
	print 'size = "16,16"'
	for grp in self.__list:
	    grp.make_graph()
	print '}'
	
    def __str__(self):
	return str(self.__list)
#End Class EltGrpList    

#class EltXMLFactory():
    #def __init__(self):

def try_getaddrinfo(addr):
    try:
	ip=IPy.IP(addr)
    except ValueError:
	try:
	    t1,t2,t3,t4,sa=socket.getaddrinfo(addr,'0',socket.AF_UNSPEC, socket.SOCK_STREAM)[0]
	    net=sa[0]
	    return net
	except socket.gaierror, e:
	    s='Error cannot resolv %s' %(addr)
	    raise ValueError(s)
    else:
	return addr
def check_protocol(elt):
    if elt.get('dport')!='' or elt.get('sport')!='':
	if elt.get('proto')!='tcp' or elt.get('proto')!='udp':
	    sys.exit('erreur proto')
	
def check_elt(elt):
    if elt.get('mark')!='' and elt.get('net')=='':
	raise ValueError('Error, element id %s has attribute mark %s defined and no net attribute'%(elt.get('ID'),elt.get('mark')))

def xml_new_elt(xmlelt):
    """Build a EltIpv4, EltNu or EltLink object from an DOM Element object
    """
    #print "entering xml_new_elt", xmlelt
    etype=xmlelt.getAttribute('type')
    ID=xmlelt.getAttribute('ID')
    name=xmlelt.getAttribute('name')
    atts=xmlelt.attributes
    e=Elt()
    h={}
    for ind in range(0,atts.length):
	kk=atts.item(ind).name
	vv=atts.item(ind).value
	e.set(kk,vv)
	h[kk]=vv
    #print "in xml_new_elt",e
    #print h
    #print h['type']
    check_elt(e)
    if e.get('net')!='':
	e.set('numnet',try_getaddrinfo(e.get('net')))
    return e
    


def xml_new_eltgrp(ID,name,op,elts):
    """Builds a EltGrp Object from an list of DOM Element objects
    """
    #print "Entering xml_new_eltgrp ", ID," ",elts
    egrp=EltGrp(ID,name,op)
    if(egrp.op=="and"):
        eltset=EltSet([])
	eltsetlink=EltSet([])
	lkl=[]
	#print "new_elt_grp elements dans la liste de longueur ",len(elts)
	#for a in elts:
	    #print e
	#print "-----------"
        for e in elts:
	   #tester quel element
	    #print "  xml element: ",e
	    myelt=xml_new_elt(e)
	    myelt.set('parent_name',name)
	    #print 'myelt',myelt
	    if(myelt.get('type')=="ipv4"): 
		#print "elt ipv4"
		eltset.add_elt(myelt)
		#print "elt added"
	    if(myelt.get('type')=="nufw"): 
		#print "elt nufw"
		eltset.add_elt(myelt)
		#print "elt added"
	    if(myelt.get('type') in ["proto","app"]): 
		#print "elt proto"
		eltset.add_elt(myelt)
		#print "elt added"
	    if(myelt.get('type')=="link"):
		#lkl.append(myelt)
		eltsetlink.add_elt(myelt)
		#egrp.add_eltlink(myelt)
	#print "fin boucle"
	if(len(eltset.elts)!=0):
	    egrp.add_eltset(eltset)
	if(eltsetlink.elts!=[]):#found links? -> add eltset links to list
		egrp.add_eltlink(eltsetlink)
	if len(eltset.elts)==1 and eltsetlink.elts==[]:#only one element: set op to <or>, for the operators algebra
	    #print 'setting op to or',egrp
	    egrp.op='or'
	    return egrp
	    
	#egrp.link_t=lkl
    elif(egrp.op=="or"):
	for e in elts:
	    myelt=xml_new_elt(e)
	    myelt.set('parent_name',name)
	    if(myelt.get('type')=="ipv4"):
		eltset=EltSet([])
		eltset.add_elt(myelt)
		egrp.add_eltset(eltset)
	    if(myelt.get('type')=="nufw"):
		eltset=EltSet([])
		eltset.add_elt(myelt)
		egrp.add_eltset(eltset)
	    if(myelt.get('type') in ["proto","app"]):
		eltset=EltSet([])
		eltset.add_elt(myelt)
		egrp.add_eltset(eltset)

	    if(myelt.get('type')=="link"):
		#print 'link',myelt,ID
	    	eltsetlink=EltSet([])
		eltsetlink.add_elt(myelt)
		egrp.add_eltlink(eltsetlink)
    return egrp
#end xml_new_eltgrp

class Rule:
    """Represents a rule from an acl
    """
    def __init__(self,src,dest,proto=None,appli=None,aclname='',aclid=-1,decision=''):
    	self.is_auto = 0
	self.__src=src
	self.__dest=dest
	self.__src.merge_elts()
	self.__dest.merge_elts()
	self.aclid=aclid
	self.decision=decision
	self.aclname=aclname
	self.netdst=None
	self.netsrc=None
	self.snetsrc=''
	self.snetdst=''
	self.cpt=0
	if proto is None:
	    self._proto=EltSet([Elt()])
	else:
	    self._proto=proto
	    self._proto.merge_elts()
	if appli is None:
	    self._appli = EltSet([Elt()])
	else:
	    self._appli = appli
	    self._appli.merge_elts()
	
    def __str__(self):
	s ='Rule %s -> %s | %s' %(self.elt_src(),self.elt_dest(),self.proto())
	return s
	
    def format(self):
	s =''
	return s
    
    def invert(self):
	#r=Rule(copy.deepcopy(self.__dest),copy.deepcopy(self.__src),proto=copy.deepcopy(self._proto),aclid=self.aclid,decision=self.decision)
	r=Rule(self.__dest,self.__src,proto=copy.deepcopy(self._proto),aclid=self.aclid,decision=self.decision)
	tmp=r.proto().get('dport')
	r.proto().set('dport',r.proto().get('sport'))
	r.proto().set('sport',tmp)
	#ntmp=copy.deepcopy(r.netdst)
	#r.netdst=copy.deepcopy(r.netsrc)
	#r.netsrc=ntmp
	r.netdst = copy.deepcopy(self.netsrc)
	r.netsrc = copy.deepcopy(self.netdst)
	r.cpt=self.cpt
	
	#tmp=r.elt_dest().get('dport')
	#r.elt_dest().set('dport',r.elt_dest().get('sport'))
	#r.elt_dest().set('sport',tmp)
	return r
    
    def oprint(self):
	print "from",self.__src
	print "to",self.__dest

    def src(self):
	return self.__src

    def dest(self):
	return self.__dest

    def elt_src(self):
	return self.__src.elts[0]

    def elt_dest(self):
	return self.__dest.elts[0]
    
    def proto(self):
	return self._proto.elts[0]
	
    def appli(self):
	return self._appli.elts[0]
   
    def set_netsrc(self,n):
	self.netsrc=n

    def set_netdst(self,n):
	self.netdst=n
    
    def from_nufw(self): #FIXME: 
	atts=['group']
	for k in self.elt_src().keys():
	    if k in atts: return True
	return False

    #FIXME FIXME: replace this function by one at acl level
    # rule is not aware of its place in the network layout
    def dst_isnot_src(self): #used by Aclcheck_dst_src_rules
	"""This function find out if a rule has its srs and its dst 
	in the same network (that is quit idiot)
	"""
	#print 'in dst_isnot_src dest |%s|' %(self.elt_dest().get('numnet'))
	#print 'in dst_isnot_src src |%s|' %(self.elt_src().get('numnet'))
	snet=self.elt_src().get('numnet')
	dnet=self.elt_dest().get('numnet')
	szone=self.elt_src().get('zone')
	dzone=self.elt_dest().get('zone')
	#if(snet=='' and szone=='internal'): return False
	#if(dnet=='' and dzone=='internal'): return False
	if(dzone=='external' or szone=='external'): return True #TODO: handle vpn zone
	ips=IPy.IP(self.elt_src().get('numnet'))
	ipd=IPy.IP(self.elt_dest().get('numnet'))
	#print "dst_is_not_src ", ips, ipd
	if ips not in ipd and ipd not in ips: 
	    #print "\ttrue"
	    return True
	#print "\tFalse"
	return False

    #def check_dst_src(self):
	#if self
	
    def __cmp__(self,other):
	return cmp(int(self.aclid),int(other.aclid))

    def get_dst_port_range(self,dir='dport'):
	s=self.proto().get(dir)
	if s.count(':')!=0:
	    return s.split(':')
	return (s,s)

    def get_src_port_range(self):
	return self.get_dst_port_range(dir='sport')
#End class Rule


class Acl:
    """Represents an acl in a NuFW acl xml file
    """
    def __init__(self,sub,ress,myfw,name='',proto=None,id='-1',decision='',appli=None):
	self.__rule_list=[]
	self._myfw=myfw
	self.id=id
	self.decision=decision
	self.__name=name
	self.cpt = 0
	self.__populate(sub,ress,proto,appli)
	self.__proto=proto
	
    def __populate(self,sub,ress,proto,appli):
	lnet=self._myfw.entreprise_nets()
        # include INTERNET (default) network in list used to create netfilter rules for authenticated acls
        if self._myfw.auth_ext and self._myfw.get_ext_net():
            lnet.append(self._myfw.get_ext_net().ip())
	#print lnet
	if appli is not None and len(appli.eltset_t) > 0:
	    fake_list_appli = appli.eltset_t
	else: fake_list_appli = [None]
	if proto  and len(proto.eltset_t) > 0:
	    fake_list_proto = proto.eltset_t
	else: fake_list_proto = [None]
	for eltsetfrom in sub.eltset_t:
	    #print 'eltsetfrom',eltsetfrom
	    for eltseto in ress.eltset_t:
		#if proto is not None:
		for eltsetproto in fake_list_proto:
		    #print 'eltsetproto',eltsetproto
		    for eltsetappli in fake_list_appli:
			l=[]
		        #print 'eltsetappli',eltsetappli
			
		        r=Rule(copy.deepcopy(eltsetfrom),copy.deepcopy(eltseto),proto=copy.deepcopy(eltsetproto),appli=copy.deepcopy(eltsetappli),
			    aclid=self.id,aclname=self.__name,decision=self.decision)
		        #print 'rule appli',r.appli()
		        #r=Rule(eltsetfrom,eltseto,proto=eltsetproto,aclid=self.id,aclname=self.__name,decision=self.decision)
		        self.cpt+=1
		        r.cpt=self.cpt
		        #print 'in acl __populate ', r
		        l.extend(filter(lambda x: x is not None, [self._make_rule(r,hfw) for hfw in lnet]))
		        #remove idiot rules that come from a net and go to the same net
		        l = self.check_dst_src_rules(l)
		        #print l
		        if(l==[]):
			    self.__rule_list.append(r)
			else: self.__rule_list.extend(l)
			
		
    
    def _make_rule(self,r,hfw): 
	res=None
	change=0
	#print 'hfw: ',hfw
	#if r.elt_src().get('numnet')=='' and not r.elt_src().get('zone') in ['external','internal','vpn']:
	if r.elt_src().get('numnet')=='':
	    #print res.elt_src().get('zone')
	    #print "hfw['net_src'] ", hfw
	    #res.elt_src().set('numnet',hfw['net_src'].ip())
	    res=copy.deepcopy(r)
	    res.elt_src().set('numnet',hfw)
	    res.elt_src().set('zone','internal')
	    change=1
	    #we dont want to generate a rule that goes to an iface and from a lan that is not the iface LAN.
	    if self._myfw.rule_inout(res) and self._myfw.rule_dst_is_src(res) is not True:
		change = 0
	if change==1: 
	    self.cpt=self.cpt+1
	    res.cpt=self.cpt
	    res.is_auto = 1
	    return res
	#print 'returning None'
	return None

    def check_dst_src_rules(self, rules_list) :
	res = []
	for r in rules_list:
	    #if r.dst_isnot_src(): 
	    if not self._myfw.rule_dst_is_src(r): 
		res.append(r)
	    elif self._myfw.rule_inout(r):
		res.append(r)
	return res

    def oprint(self):
	for r in self.__rule_list:
	    r.oprint()
	    print
	    
    def rules(self):
	for r in self.__rule_list:
	    yield r
    def fw(self):
	return self._myfw

def parse_groups(doc):
    l = {}
    groups = doc.getElementsByTagName('groups')
    groups = groups[0]
    for group in groups.getElementsByTagName('group'):
        id = int(group.getAttribute('ID'))
        enabled = group.getAttribute('enabled')
        st = 0
        if enabled == '1':
            st = 1
        l[id] = st
    return l
