#
# Copyright(C) 2005 INL
# Written by Jean Gillaux <jean@inl.fr>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#


"""Common stuff to nufw acls backends.

"""

__version__ = '0.01'
__author__ = 'Jean Gillaux'

from sets import Set
from IPy import IP
import nuacl

A_SRC_IP_START='SrcIPStart'
A_SRC_IP_END='SrcIPEnd'
A_DST_IP_START='DstIPStart'
A_DST_IP_END='DstIPEnd'

A_SRC_PORT_START='SrcPortStart'
A_SRC_PORT_END='SrcPortEnd'
A_DST_PORT_START='DstPortStart'
A_DST_PORT_END='DstPortEnd'
A_SRC_PORT='SrcPort'
A_DST_PORT='DstPort'

A_PROTO='Proto'
A_DECISION='Decision'
A_GROUP='Group'

A_OS_NAME='OsName'
A_OS_VERSION='OsVersion'
A_OS_RELEASE='OsRelease'

A_APP_NAME='AppName'
A_APP_SIG='AppSig'

A_DESC='description'


ATTR_TYPE_MULTIPLE=Set([A_SRC_PORT,A_DST_PORT,A_PROTO,A_GROUP,A_OS_NAME,A_OS_VERSION,A_OS_RELEASE,A_APP_NAME,A_APP_SIG])
ATTR_TYPE_SIMPLE=Set([A_SRC_IP_START,A_SRC_IP_END,A_DST_IP_START,A_DST_IP_END,A_SRC_PORT_START,A_SRC_PORT_END,A_DST_PORT_START,A_DST_PORT_END,A_DECISION,A_DESC])
ALL_ATTR=(ATTR_TYPE_MULTIPLE | ATTR_TYPE_SIMPLE) 
COMPARE_ATTR=ALL_ATTR

backend_encoding='utf-8'

class NURule(object):
    def __init__(self,xmlrule=None):
	self.h = NURule._init_h()
	self.__rcpt = 0
	self.name=''
	self.ATTR_TYPE_SIMPLE = ATTR_TYPE_SIMPLE
	self.ATTR_TYPE_MULTIPLE = ATTR_TYPE_MULTIPLE
	self.COMPARE_ATTR = COMPARE_ATTR
	self.xmlrule=xmlrule
	if xmlrule is not None:
	    self.nurule_from_xml_rule(xmlrule)
	    
    def _init_h(self):
	"""Initialization.
	"""
	res={}
	#print 'ALL_ATTR:',ALL_ATTR
	for item in ALL_ATTR:
	    res[item]=Set()
	return res
    _init_h=classmethod(_init_h)
   
    def _update_hash(self):
	pass

    def make_decision(self,dec):
	if dec == 'drop' or dec =='reject': return '0'
	return '1' #default 1
    make_decision = classmethod(make_decision)
    
    def nurule_from_xml_rule(self,r,dn=''):
	"""Set NURule attributes from xml rule object r.
	"""
	appli = r.appli()
	res = self
	nsrc = IP(r.elt_src().get('numnet'))
	ndst = IP(r.elt_dest().get('numnet'))
	res[A_SRC_IP_START] = nsrc.net().strDec()
        res[A_SRC_IP_END] = nsrc.broadcast().strDec()
	res[A_DST_IP_START] = ndst.net().strDec()
	res[A_DST_IP_END] = ndst.broadcast().strDec()
        res[A_DECISION] = NURule.make_decision(r.decision)
	res[A_GROUP] = r.elt_src().get('group').encode("utf-8")
        res[A_PROTO] = str(nuacl.proto_number(r.proto().get('proto')))
	src_port_start, src_port_end = r.get_src_port_range()
	if src_port_start != '':res[A_SRC_PORT_START]=src_port_start.encode(backend_encoding)
	if src_port_end != '':res[A_SRC_PORT_END]=src_port_end.encode(backend_encoding)
        #print 'port src',res[A_SRC_PORT_START],res[A_SRC_PORT_END]
	dst_port_start, dst_port_end = r.get_dst_port_range()
	if dst_port_start == '': dst_port_start='0'
	if dst_port_end == '': dst_port_end = '65536'
	res[A_DST_PORT_START],res[A_DST_PORT_END] = (dst_port_start, dst_port_end)
	res[A_DST_PORT] = dst_port_end
	#ldap description is limited to 1024 characters.
        res[A_DESC] = r.aclname.encode("utf-8")[0:1023] 
	res.set_if_not_empty(A_OS_NAME,appli.get('OSName'))
	res.set_if_not_empty(A_OS_RELEASE,appli.get('OSRelease'))
	res.set_if_not_empty(A_OS_VERSION,appli.get('OSVersion'))
	res.set_if_not_empty(A_APP_NAME,appli.get('ApplicationPath'))
	
    #nurule_from_xml_rule = classmethod(nurule_from_xml_rule)
    
    def __setitem__(self,key,value):
	#if not key in ALL_ATTR: raise ValueError('bad ldap attribute %s in __setitem__' %(key))
	if key in self.ATTR_TYPE_MULTIPLE:
	    #print 'in __setitem__ value=',value
	    if type(value)==list: self.h[key]=Set(value)
	    else: self.h[key]=Set([value])
	if key in self.ATTR_TYPE_SIMPLE:
	    if type(value)==list: 
		if len(value)>0: self.h[key]=Set([value[0]])
		else: self.h[key]=Set([])
	    else: self.h[key]=Set([value])

    def __getitem__(self,key):
	if not self.h.has_key(key): raise ValueError('bad ldap attribute %s in __getitem__' %(key))
	return [i for i in self.h[key]]
    
    def set_if_not_empty(self,attr,value):
	"""set attr to value if value is not empty char
	"""
	if value != '': self[attr] = value

    def __eq__(self,other):
	#print 'in __eq__ CN',self.h[A_CN],other.__h[A_CN]
	#print 'other.h',other.h,type(other)
	for k,v in self.h.iteritems():
	    if k in self.COMPARE_ATTR:
		if other.h[k]!=v: 
		    #print '__eq__ return False'
		    return False
	#print '__eq__ return True'
	return True

    def __ne__(self,other):
	return not self.__eq__(other)

    def __str__(self):
	s='NURule ('
	for k,v in self.h.iteritems():
	    s+=str(k)+'->'+str(v)+', '
	s+=')'
	return s


class NUBackend(object):
    """Parent class for nufw acls backends.
    """
    
    def __init__(self,fw=None,backend_class=None):
	self.rules = []
	self.__rcpt = 0
	self.backend_class = NURule
	# :@)
	if backend_class is not None: self.backend_class = backend_class
	if fw is not None:
	    for r in fw.allrules():
		if r.from_nufw():
		    if r.proto().get('proto') == 'tcp':
			if r.proto().get('decision') not in ['log', 'ulog']:
			    self.add_rule(self.backend_class(xmlrule=r))
    
    def add_rule(self,rule):
	self.rules.append(rule)
	
    def set_rules(self,list_of_rules):
	self.rules=list_of_rules
 
    def get_rules(self):
	return self.rules
	
    def process(self):
	pass
